import java.applet.Applet;
import java.awt.Color;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.image.ImageObserver;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;
import java.util.StringTokenizer;

/**
 * This is a simple Text Scrolling Applet
 * Author: Calvin
 * Date: 19/6/99
 * Last updated: 19/9/99
 */

public class ScrollApplet extends Applet implements Runnable
{
    private final int x = 10;                   // left margin
    private int offset;                         // Y offset of the messages
    private int interval;                       // Interval for thread to sleep
    private int separate;                       // Separation between each message
    private boolean underline = false;          // Whether the bgimage is loaded or not
    private String[] messages = new String[0];  // Messages to display
    private URL[] url = new URL[0];             // The URL
    private URL default_url;                    // The default URL for each message
    private Color link;                         // Color on mouse over
    private Font font;                          // The font
    private Thread thread;                      // The Thread to be run
    private Image bgimage = null;               // Background image
    private Image image;                        // The image created
    private Graphics g;                         // The Graphics object for painting

    /**
     * Initilize the parameter and applet
     */
    public void init()
    {
        // Get normal text color
        if(getParameter("textcolor1") != null)
            setForeground(toColor(getParameter("textcolor1")));
        else
            setForeground(Color.black);

        // Get background color
        if(getParameter("background") != null)
            setBackground(toColor(getParameter("background")));
        else
            setBackground(Color.white);

        // Get color when mouse over
        if(getParameter("textcolor2") != null)
            link = toColor(getParameter("textcolor2"));
        else
            link = Color.blue;

        // Get default URL for all messages
        if(getParameter("url") != null)
            default_url = toURL(getParameter("url"));
        else
            default_url = null;

        // Get the time interval
        if(getParameter("speed") != null)
        {
            try { interval = Integer.parseInt(getParameter("speed")); }
            catch(NumberFormatException e) { interval = 10; }
        } else
            interval = 10;

        // Get separation between each message
        if(getParameter("separate") != null)
        {
            try
            {
                separate = Integer.parseInt(getParameter("separate"));
                if(separate < 0)
                    separate = 40;
            } catch(NumberFormatException e) { separate = 40; }
        } else
            separate = 40;

        if(getParameter("underline") != null)
        {
            String temp = getParameter("underline").toUpperCase();
            if(temp.equals("TRUE"))
                underline = true;
        }

        // Get font size and initilize the Font
        if(getParameter("font") != null)
        {
            try
            {
                font = new Font("Dialog", Font.PLAIN, Integer.parseInt(getParameter("font")));
            } catch(NumberFormatException e)
            {
                font = getFont();
            }
        } else
        {
            font = getFont();
        }

        getMessages();
        getURL();
        getBackgroundImage();

        offset = 0 - size().height;
        setFont(font);                                               // Set font
        image = createImage(size().width, size().height);            // Offscreen image
        g = image.getGraphics();                                     // Graphics for painting
    }

    /**
     * Start the thread
     */
    public void start()
    {
        if(thread == null || !thread.isAlive())
        {
            thread = new Thread(this);
            thread.start();
        }
    }

    /**
     * Stop the running thread
     */
    public void stop()
    {
        if(thread != null && thread.isAlive())
        {
            thread.stop();
            thread = null;
        }
    }

    /**
     * Suspend the running thread
     */
    public boolean mouseEnter(Event evt, int x, int y)
    {
        if(thread != null)
        {
            thread.suspend();
            int num = getIndex(offset + y);
            if(num != -1)
            {
                if(url[num] != null)
                    showStatuText(num);
                else if(default_url != null)
                    showStatuText(-1);
                paintImage(num);
                repaint();
            }
        }
        return true;
    }

    /**
     * Resume the suspended thread
     */
    public boolean mouseExit(Event evt, int x, int y)
    {
        showStatus("");
        if(thread != null)
        {
            thread.resume();
            paintImage(-1);
            repaint();
        }
        return true;
    }

    /**
     * Handle the mouse down event
     */
    public boolean mouseDown(Event evt, int x, int y)
    {
        int index = getIndex(offset + y);
        if(index != -1)
            openURL(index);
        return true;
    }

    /**
     * Handle the mouse move event
     */
    public boolean mouseMove(Event evt, int x, int y)
    {
        int num = getIndex(offset + y);
        if(num != -1)
        {
            if(url[num] != null)
                showStatuText(num);
            else if(default_url != null)
                showStatuText(-1);
            else
                showStatus("");
        } else
            showStatus("");
        paintImage(num);
        repaint();
        return true;
    }

    /**
     * Content of thread
     */
    public void run()
    {
        offset = 0 - size().height;
        while(thread != null)
        {
            try { thread.sleep(interval); }
            catch(InterruptedException e) { System.out.println("Error"); }
            paintImage(-1);
            repaint();
            if(offset >= getMessageHeight(messages.length - 1))
                offset = 0;
            else
                offset++;
        }
    }

    /**
     * Overrided to prevent flicker
     */
    public void update(Graphics p)
    {
        paint(p);
    }

    /**
     * Draw the image to the screen
     */
    public void paint(Graphics p)
    {
        p.drawImage(image, 0, 0, null);
    }

    /*
     * Open a url
     */
    private void openURL(int index)
    {
        if(url[index] == null && default_url == null)
            return;
        if(url[index] != null)
            getAppletContext().showDocument(url[index], "_self");
        else
            getAppletContext().showDocument(default_url, "_self");
    }

    /*
     * Convert a string to a URL
     */
    private URL toURL(String str)
    {
        URL address = null;
        if(str.indexOf("://") == -1)
        {
            URL temp = getDocumentBase();
            String filename;
            if(str.charAt(0) == '/')
                filename = "";
            else
                filename = temp.getFile();

            int lastindex = filename.lastIndexOf("/"); 
            try
            {
                address = new URL(temp.getProtocol(),
                          temp.getHost(),
                          filename.substring(0, lastindex + 1) + str);
            } catch(MalformedURLException e) { System.out.println("Error"); }
        } else
        {
            try { address = new URL(str); }
            catch(MalformedURLException e) { System.out.println("Error"); }
        }
        return address;
    }

    /*
     * Paint the offscreen image
     */
    private void paintImage(int highlight)
    {
        FontMetrics fm = getFontMetrics(font);
        int font_height = fm.getHeight();
        int y, num_lines;
        int index = getMessageIndex(offset);
        if(index != 0)
            y = getMessageHeight(index - 1) - offset + font_height;
        else
            y = 0 - offset + font_height;

        g.clearRect(0, 0, size().width, size().height);
        if(bgimage != null)
            g.drawImage(bgimage, 0, 0, size().width, size().height, null);

        String line;
        while(y < size().height + 1)
        {
            String temp = messages[index];
            num_lines = getNumberOfLines(temp);
            if(index == highlight)
                g.setColor(link);
            for(int i = 0; i < num_lines; i++)
            {
                line = getLine(temp, i);
                g.drawString(line, x, y);
                y++;
                if(underline)
                    g.drawLine(x, y, fm.stringWidth(line) + x, y);
                y += font_height;
            }
            if(index == highlight)
                g.setColor(getForeground());
            y += separate;
            if(index == messages.length -1)
                index = 0;
            else
                index++;
        }
    }

    /*
     * Get the messages parameter
     */
    private void getMessages()
    {
        String temp;
        Vector v = new Vector();
        FontMetrics fm = getFontMetrics(font);
        int i = 0;
        while((temp = getParameter("message" + i)) != null)
        {
            v.addElement(strFormat(temp, size().width-2*x, fm));
            i++;
        }
        messages = new String[v.size()];
        v.copyInto(messages);
    }

    /*
     * Get the URL for each messages
     */
    private void getURL()
    {
        url = new URL[messages.length];
        for(int j = 0; j < messages.length; j++)
        {
            if(getParameter("url" + j) != null)
                url[j] = toURL(getParameter("url" + j));
            else
                url[j] = null;
        }

    }

    /*
     * Download the background image
     */
    private void getBackgroundImage()
    {
        if(getParameter("bgimage") != null)
        {
            showStatus("Loading Image ...... ");
            bgimage = getImage(getDocumentBase(), getParameter("bgimage"));
            MediaTracker mt = new MediaTracker(this);
            mt.addImage(bgimage, 0);
            try { mt.waitForAll(); }
            catch(InterruptedException e) { }
            showStatus("");
        }
    }

    /*
     * Get the height of messages up to index
     * @param msg_index The index of the message
     */
    private int getMessageHeight(int msg_index)
    {
        int total_lines = 0, index = 0, num;
        String str;
        for(int i = 0; i < msg_index + 1; i++)
        {
            str = messages[i];
            total_lines += getNumberOfLines(str);
        }
        num = (msg_index + 1)*separate + total_lines*getFontMetrics(font).getHeight() + total_lines;
        return num;
    }

    /*
     * Get the index of the message according to the input y offset
     * Return -1 if the y offset is in the area between two messages.
     */
    private int getIndex(int y_offset)
    {
        if(y_offset < 0)
            return -1;
        if(y_offset > getMessageHeight(messages.length - 1))
            y_offset -= getMessageHeight(messages.length - 1);
        int index = getMessageIndex(y_offset);
        int messages_height = getMessageHeight(index);
        if(messages_height - y_offset < separate)
            return -1;
        return index;
    }

    /*
     * Get the index of a message according to the input y offset.
     * Even the y offset is in the area between two messages,
     * it returns the index of message which just above that blank area.
     */
    private int getMessageIndex(int y_offset)
    {
        FontMetrics fm = getFontMetrics(font);
        int index = 0, y_location = 0;
        while(true)
        {
            y_location = getMessageHeight(index);
            if(y_location >= y_offset)
                break;
            index++;
        }
        return index;
    }

    /*
     * Count the number of lines in the input string
     * @param str The string to be counted
     */
    private int getNumberOfLines(String str)
    {
        int index = 0, temp, num_of_lines = 0;
        while((temp = str.substring(index).indexOf("\n")) != -1)
        {
            num_of_lines++;
            index += temp + 1;
        }
        if(str.substring(index).length() > 0)
            num_of_lines++;
        return num_of_lines;
    }

    /*
     * Get the string and return a specific line from the input string
     */
    private String getLine(String str, int line_num)
    {
        if(line_num > getNumberOfLines(str) - 1)
            return "";
        int index = 0, end = 0;
        if(str.indexOf("\n") != -1)
            end = str.indexOf("\n");
        else
            end = str.length();
        for(int i = 0; i < line_num; i++)
        {
            index += end + 1;
            end = str.substring(index).indexOf("\n");
            if(end == -1)
                end = str.substring(index).length();
        }
        return str.substring(index, index + end);
    }

    /*
     * Convert a string with format 0, 0, 0 to color and return
     * @param color The input string
     */
    private Color toColor(String color)
    {
        StringTokenizer st = new StringTokenizer(color, " ,");
        int red, blue, green;
        red = Integer.parseInt(st.nextToken());
        blue = Integer.parseInt(st.nextToken());
        green = Integer.parseInt(st.nextToken());
        return new Color(red, blue, green);
    }

    /*
     * Show URL in the status bar
     */
    private void showStatuText(int i)
    {
        if(i == -1)
            showStatus(default_url.toString());
        else
            showStatus(url[i].toString());
    }

    /*
     * Format a large string to fit with a given width and font metrics.
     * each line is separated by a '\n'
     */
    private String strFormat(String str, int width, FontMetrics fm)
    {
        int firstindex = 0, lastindex = 0, length = 0;
        StringBuffer buffer = new StringBuffer("");
        if(str.charAt(str.length() - 1) != '\n')
            str = str.concat("\n");

        while((lastindex = str.substring(firstindex).indexOf('\n')) != -1)
        {
            length = fm.stringWidth(str.substring(firstindex, firstindex+lastindex));
            if(length <= width)
            {
                buffer.append(str.substring(firstindex, firstindex+lastindex));
                buffer.append('\n');
            } else
            {
                buffer.append(splitLine(
                        str.substring(firstindex, firstindex+lastindex), width, fm));
                buffer.append('\n');
            }
            firstindex += lastindex + 1;
        }
        return buffer.toString();
    }    // Method strFormat end

    /*
     * Split a line to fit a given width and font metrics
     */
    private String splitLine(String line, int width, FontMetrics fm)
    {
        StringBuffer buffer = new StringBuffer("");
        int firstindex = 0, lastindex = 0, length = 0;
        String space = " ", temp = "";
        if(line.charAt(line.length() - 1) != ' ')
            line = line.concat(space);
        while((lastindex = line.substring(firstindex).indexOf(space)) != -1)
        {
            length = fm.stringWidth(temp) +
                 fm.stringWidth(line.substring(firstindex, firstindex+lastindex));
            if(length <= width)
            {
                temp = temp.concat(line.substring(firstindex, firstindex+lastindex)).concat(space);
            } else
            {
                buffer.append(temp.trim()).append('\n');
                temp = "";
                temp = temp.concat(line.substring(firstindex, firstindex+lastindex));
                length = fm.stringWidth(temp);
                if(length > width)
                {
                    temp = splitWord(buffer, temp, width, fm);
                }
                if(temp.charAt(temp.length() - 1) != ' ')
                    temp = temp.concat(space);
            }
            firstindex += lastindex + 1;
        }
        if(temp.length() > 0)
            buffer.append(temp.trim());

        return buffer.toString();
    }

    /*
     * Split a continues word to fit a given width and font metrics
     */
    private String splitWord(StringBuffer buffer, String word, int width, FontMetrics fm)
    {
        char[] chars = word.toCharArray();
        int offset = 0, end = 1, length = 0;
        for(; (offset+end) < chars.length; end++)
        {
            length = fm.charsWidth(chars, offset, end);
            if(length > width)
            {
                buffer.append(chars, offset, end - 1).append('\n');
                end--;
                offset += end;
                end = 0;
            }
        }
        return new String(chars, offset, end);
    }
}
